/*
 * A simple program to transfer CD data to DAT
 *
 * PLEASE NOTE!  I (Charles Henrich) have had no luck in using this code when
 * mediad is monitoring the dat devices.  If you encounter strange problems
 * disable mediad and try again.
 *
 * Doug Cook
 * Silicon Graphics, Inc., December 1993
 *
 * Charles Henrich (henrich@msu.edu), December 1994
 *
 *   - Append mode
 *   - Start/End Track
 *   - Play while Copy
 *   - TrackList Copy
 *   - Catalog Number Copy
 *   - Configuration File Support
 *
 * These portions (c) Copyright 1994, Charles R. Henrich, ALL RIGHTS RESERVED,
 * all others:
 *
 * (c) Copyright 1993, Silicon Graphics, Inc.
 * ALL RIGHTS RESERVED
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the above
 * copyright notice appear in all copies and that both the copyright notice
 * and this permission notice appear in supporting documentation, and that
 * the name of Silicon Graphics, Inc. (or Charles Henrich) not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.
 *
 * THE MATERIAL EMBODIED ON THIS SOFTWARE IS PROVIDED TO YOU "AS-IS"
 * AND WITHOUT WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR OTHERWISE,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF MERCHANTABILITY OR
 * FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL SILICON
 * GRAPHICS, INC. (OR CHARLES HENRICH) BE LIABLE TO YOU OR ANYONE ELSE
 * FOR ANY DIRECT, SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES
 * OF ANY KIND, OR ANY DAMAGES WHATSOEVER, INCLUDING WITHOUT LIMITATION,
 * LOSS OF PROFIT, LOSS OF USE, SAVINGS OR REVENUE, OR THE CLAIMS OF
 * THIRD PARTIES, WHETHER OR NOT SILICON GRAPHICS, INC. (OR CHARLES HENRICH)
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH LOSS, HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * POSSESSION, USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * US Government Users Restricted Rights
 * Use, duplication, or disclosure by the Government is subject to
 * restrictions set forth in FAR 52.227.19(c)(2) or subparagraph
 * (c)(1)(ii) of the Rights in Technical Data and Computer Software
 * clause at DFARS 252.227-7013 and/or in similar or successor
 * clauses in the FAR or the DOD or NASA FAR Supplement.
 * Unpublished-- rights reserved under the copyright laws of the
 * United States.  Contractor/manufacturer is Silicon Graphics,
 * Inc., 2011 N.  Shoreline Blvd., Mountain View, CA 94039-7311.
 *
 */


#include <sys/types.h>
#include <sys/errno.h>
#include <sys/tpsc.h>
#include <sys/mtio.h>
#include <sys/prctl.h>
#include <fcntl.h>
#include <cdaudio.h>
#include <dataudio.h>
#include <signal.h>
#include <stdio.h>
#include <audio.h>

/*
 * #define AUDIO if you want to play the CD to the audio system
 * instead of writing to DAT.
 */
#define NFRAMES 12

#define QUEUESIZE 68
#define HIGHWATER 34

static char *datdev = "/dev/nrtape";

static CDPLAYER *cd;
static int datfd;            /* dat tape fd */
static int verbose = 1;
static int aptrack = 299;
static char *configfile = NULL;
static int tracklistlen = -1;
static int tracklistcount = 0;
static int tracklist[99] = {'\0'};
static int append = 0;
static int program = 0;
static int starttrack = -1;
static int endtrack = -1;
static int abortrecording = -1;
static int highprog = 0;
static int playaudio = 0;
struct mtaudtimecode last_valid_atime;

static int startidframes;    /* # of frames left in DAT track start ID */
static ushort playbuf[DTDA_NUMSAMPS44K*2];

/*
 * we define a word-aligned DAT frame. Writes directly to a scsi
 * device can only come from word-aligned addresses. If we made an
 * array of regular DAT frames, every other one would not be word-aligned.
 */
typedef struct wadtframe {        /* word-aligned DAT frame */
    DTFRAME frame;
    short pad;
} WADTFRAME;


/*
 * We implement a circular queue of DAT audio frames. One process
 * reads from the CD and writes data to the queue in DAT format.
 * when the size of the queue hits a high-water mark, another process
 * wakes up and begins to write data to the DAT from the other end
 * of the queue. This extra buffering is required because both CD and
 * DAT run at "exactly" 44.1kHz, but off of different clocks. They
 * may therefore generate or consume data at slightly different rates.
 */

static WADTFRAME dtq[QUEUESIZE];    /* queue of DAT audio frames */
static volatile int head = 0, tail = 0;    /* queue head & tail pointers */
static volatile int finish = 0;

/*
 * dtf is the "template" DAT frame. It is modified continually by the
 * process reading the CD, then copied into the queue. This prevents
 * us from having to update all the DAT fields all the time; we just
 * update those fields which change from frame to frame.
 */
static DTFRAME dtf;
static int dtf_nsamps=0;    /* number of audio samps currently in dtf */
static DTFRAME empty;
static index;             /* current index being recorded */

void stopdat(int sig)
{
    static int beenhere = 0;
    struct mtop mtc;

    finish = 1;
    signal(SIGTERM, SIG_IGN);
    signal(SIGINT, SIG_IGN);
    signal(SIGKILL, SIG_IGN);

    if(beenhere == 1) exit();
    beenhere = 1;

    fprintf(stderr,"Aborting tape operation\n");

    ioctl(datfd, MTABORT);
    exit(0);
}


void
swapcopy(ushort *a, ushort *b, int n)
{
    ushort *done = a + n;
    while(a != done) {
    /* DAT audio frames must have the audio byte-swapped */
        *b++ = ((*a & 0xff) << 8)+(*a >> 8);
        a++;
    }
}


void
cd_cat_func(arg, type, catno)
void *arg; CDDATATYPES type; unchar *catno;
{
    struct dtcatalogpack *dcp;
    int count;

    dcp = (struct dtcatalogpack *)&dtf.subcode.packs[DTP_CATALOG-1];

    bzero(dcp,sizeof(struct dtcatalogpack));

    fprintf(stderr,"Catalog number:");
    for(count=0;count<13;count++) fprintf(stderr,"%x",catno[count]);
    fprintf(stderr,"\n");
    dcp->id = DTP_CATALOG;
    dcp->n1 = catno[0];
    dcp->nrest[0].dhi = catno[1];
    dcp->nrest[0].dlo = catno[2];
    dcp->nrest[1].dhi = catno[3];
    dcp->nrest[1].dlo = catno[4];
    dcp->nrest[2].dhi = catno[5];
    dcp->nrest[2].dlo = catno[6];
    dcp->nrest[3].dhi = catno[7];
    dcp->nrest[3].dlo = catno[8];
    dcp->nrest[4].dhi = catno[9];
    dcp->nrest[4].dlo = catno[10];
    dcp->nrest[5].dhi = catno[11];
    dcp->nrest[5].dlo = catno[12];
}

void
cd_control_func(arg, type, flags)
void *arg; CDDATATYPES type; unsigned char *flags;
{
    /*
     * Determine whether or not copy protect is enabled
     * on the CD. If so, we must pass it through to the
     * DAT.
     */
    dtf.subcode.mid.copy =
        ((*flags & CDQ_COPY_MASK) == CDQ_COPY_PERMITTED) ?
         DTM_COPY_PERMITTED : DTM_COPY_PROHIBITED;
}

void
cd_index_func(void *arg, CDDATATYPES type, CDPROGNUM *inum)
{
    struct dttimepack *tpp;
    struct dttimecode *tcp;

    index = inum->value;

    if (verbose) {
        printf("\nindex %d\n",index);
    }

    /*
     * set up the subcode bits
     */
    tpp = (struct dttimepack *)&dtf.subcode.packs[DTP_ATIME-1];
    tpp->index.dhi = index / 10;
    tpp->index.dlo = index % 10;
    tpp = (struct dttimepack *)&dtf.subcode.packs[DTP_PTIME-1];
    tcp = &tpp->tc;
    tpp->index.dhi = index / 10;
    tpp->index.dlo = index % 10;
    if (index == 0) {
    /*
     * The DAT spec states that when index=0, times stay
     * at 0.
     */
    bzero(tcp,sizeof(struct dttimecode));
    }
    /* ...and we don't bother with r-time */
}


void
cd_pnum_func(void *arg, CDDATATYPES type, CDPROGNUM *pnum)
{
    int p = pnum->value;
    struct dttimepack *tpp;
    struct dttimecode *tcp;
    static int seektrack = 0;

    if(tracklistlen != -1)
        {
        if(p != seektrack)
            {
            if(verbose)
                {
                fprintf(stderr,"Seeking to CD track %d\n",
                        tracklist[tracklistcount]);
                }

            while(CDseektrack(cd, tracklist[tracklistcount]) == -1)
                {
                tracklistcount++;
                if(tracklistcount >= tracklistlen)
                    {
                    abortrecording = 1;
                    return;
                    }
                }

            seektrack=p=tracklist[tracklistcount];

            tracklistcount++;
            }
        }

/* We want to start from track 1 even if the CD is track 10 */
/* We want to start from the appended track #, also ignoring CD track */

    if(endtrack != -1)
        {
        if(p > endtrack)
            {
            abortrecording = 1;
            if(verbose) printf("\nReached end track\n");
            return;
            }
        }

    if (verbose) {
        printf("\n\nCD Program %d ",p);
    }

    if(tracklistlen != -1)
        {
        p = highprog + tracklistcount;
        }
    else
        {
        if(starttrack != -1) p+=(highprog-starttrack);
        else p+=highprog;
        }

    if (verbose) {
        printf(" DAT Program %d\n",p);
    }
    /*
     * set up the subcode bits
     */
    tpp = (struct dttimepack *)&dtf.subcode.packs[DTP_ATIME-1];

    tpp->pno2 = dtf.subcode.sid.pno2 = p / 10;
    tpp->pno3 = dtf.subcode.sid.pno3 = p % 10;
    tpp = (struct dttimepack *)&dtf.subcode.packs[DTP_PTIME-1];
    tpp->pno2 = dtf.subcode.sid.pno2;
    tpp->pno3 = dtf.subcode.sid.pno3;
    /* ...and we don't bother with r-time */

    tcp = &tpp->tc;

    /* don't bother writing hours since they're always 0 */
    tcp->mhi = 0;
    tcp->mlo = 0;
    tcp->shi = 0;
    tcp->slo = 0;
    tcp->fhi = 0;
    tcp->flo = 0;

    /*
     * now we set it up so that the next 300 frames
     * have the start id bit set, indicating start-of-program.
     * DAT spec sez START must be set for 300 frames at start-of-program
     * and PRIORITYID must be set whenever program # is valid.
     */
    dtf.subcode.sid.ctrlid |= (DTS_PRIORITYID | DTS_START);
    startidframes = 300;
}

void
write_DAT(void *fd)
{
    int filled;
    int framecount=0;
    struct dttimepack *tpp;
    struct dttimecode *tcp;
    static ALport p = 0;
    ALconfig c;

    do {
        filled = tail - head;
        if (filled < 0) {
            filled += QUEUESIZE;
        }
    } while (filled < HIGHWATER);

    while (1) {
        while (tail == head) {
            if (finish)
                {
                fprintf(stderr,"\n");
                return;
                }
        }
        if (write((int)fd,&(dtq[head].frame),sizeof(DTFRAME)) < 0) {
            perror("DAT write failed");
            exit(-1);
        }

        framecount++;
        if(verbose) fprintf(stderr,"Writing frame %d\r",framecount);

        if(playaudio == 1)
            {
            if (!p)
                {
                c = ALnewconfig();
                ALsetqueuesize(c,DTDA_NUMSAMPS44K*16);
                p = ALopenport("CDtest","w",c);
                ALfreeconfig(c);
                }

           swapcopy((ushort *)&dtq[head].frame.audio,playbuf,DTDA_NUMSAMPS44K);
            ALwritesamps(p, playbuf, DTDA_NUMSAMPS44K);
            }

    /* we increment head carefully so it only takes 1 write to do so */
        if (head == QUEUESIZE-1) {
        head = 0;
    }
    else {
        head++;
    }
    }
}

void
fix_parity()
{
    volatile unchar *pp;
    int i;

    /*
     * calculate the parity bytes for the packs we are using.
     */
    for(i = 0; i<dtf.subcode.sid.numpacks; i++) {
        pp = (volatile unchar *)&dtf.subcode.packs[i];
        dtf.subcode.packs[i].parity
            = pp[0] ^ pp[1] ^ pp[2] ^ pp[3] ^ pp[4] ^ pp[5] ^ pp[6];
    }
}

void
adjust_time()
{
    struct dttimepack *tpp;
    struct dttimecode *tcp;

    /*
     * increment Atime normally.
     */
    tpp = (struct dttimepack *)&dtf.subcode.packs[DTP_ATIME-1];
    tcp = &tpp->tc;
    DTinctime(tcp);

    /*
     * The DAT spec states that when index=0, ptime stays
     * at 0. Otherwise, we increment the ptime.
     */
    if (index != 0) {
        tpp = (struct dttimepack *)&dtf.subcode.packs[DTP_PTIME-1];
        tcp = &tpp->tc;
        DTinctime(tcp);
    }

    if (startidframes > 0) {
        startidframes--;
        if(startidframes==0) {
            /* shut off start ID bit after startidframes */
            dtf.subcode.sid.ctrlid  &= ~DTS_START;
        }
    }
}


void
cd_audio_func(void *arg, CDDATATYPES type, short *audio)
{
    int filled;
    int qfull = 0;

    /*
     * The size of a CD frame and the size of a DAT frame vastly differ.
     * The DAT frames are much larger. We have to accumulate enough CD
     * frames to fill a DAT frame before we can stick the DAT frame on
     * the queue to be written. There are two cases: (1) the CD frame
     * fills the DAT frame, in which case it must be split into two
     * pieces, the full DAT frame shoved onto the queue, and the remaining
     * piece used to partially fill the next DAT frame; and (2) the entire CD
     * frame fits into the DAT frame, in which case we just stick it in.
     */
    if (dtf_nsamps + CDDA_NUMSAMPLES >= DTDA_NUMSAMPS44K) {
    /*
     * case 1: we must split the CD frame and write the DAT frame
     * s is the split point.
     */
        int s = DTDA_NUMSAMPS44K - dtf_nsamps;
        swapcopy((ushort*)audio,((ushort *)&dtf.audio)+dtf_nsamps, s);

        /* if queue is full, wait until space is available */
        do {
            filled = tail - head;
            if (filled < 0) {
                filled += QUEUESIZE;
            }

            if (filled == QUEUESIZE - 1 && qfull==0) {
                printf("queue is full! DAT may have glitches.\n");
        qfull = 1;
        }

        } while (filled >= QUEUESIZE - 1);
    qfull = 0;

    /*
     * calculate the parity in the subcode packs, write the DAT
     * frame to the queue, and adjust the times in the subcode packs.
     */
        fix_parity();
        bcopy(&dtf,&dtq[tail].frame,sizeof(DTFRAME));
        adjust_time();

    /* we increment tail carefully so it only takes 1 write to do so */
        if (tail == QUEUESIZE-1) {
        tail = 0;
    }
    else {
        tail++;
    }

        swapcopy((ushort *)audio+s, (ushort *)&dtf.audio, (CDDA_NUMSAMPLES-s));
        dtf_nsamps = CDDA_NUMSAMPLES-s;
    }
    else {
    /*
     * case 2: the CD frame fits in the DAT frame
     */
        swapcopy((ushort *)audio,((ushort *)&dtf.audio)+dtf_nsamps, CDDA_NUMSAMPLES);
        dtf_nsamps += CDDA_NUMSAMPLES;
    }
}

int
get_firmware_revision(int fd, int *maj, int *min)
{
    ct_g0inq_data_t info;
    char revbuf1[MAX_INQ_PRL + 1];

    if (ioctl(fd, MTSCSIINQ, &info) >= 0) {
        strncpy(revbuf1, (char *)info.id_prl, MAX_INQ_PRL);
        *maj = atoi(strtok(revbuf1, "."));
        *min = atoi(strtok(NULL, "."));
        return 0;
    }
    else {
        return -1;     /*ioctl failed */
    }
}

int
finish_dat(int fd)
{
    int i;
    struct mtop mtc;

    /*
     * use the same dummy frame we used to write the leader,
     * except switch from lead-in to lead-out subcodes.
     */
    empty.subcode.sid.pno1 = 0;
    empty.subcode.sid.pno2 = empty.subcode.sid.pno3 = 0xE;   /* EOT */
    empty.subcode.sid.ctrlid =  DTS_START;

    /*
     * Write this dummy frame to the tape to make the trailer
     * DAT spec sez lead-out must be >=300 frames
     */

    fprintf(stderr,"Writing leadout\n");

    for (i = 0; i < 300; i++) {
        if ((write(fd, &empty, sizeof(DTFRAME))) < 0) {
            perror("tape write failed on trailer");
            return 0;
        }
    }

    fprintf(stderr,"Rewinding tape\n");
    mtc.mt_op = MTREW;
    mtc.mt_count = 1;
    ioctl(fd, MTIOCTOP, &mtc);
}

int
prepare_dat(int fd)
{
    int i;
    struct mtop mtc;
    struct mtaudio mta;
    struct mtget mtg;
    struct dttimecode *tcp;
    struct dttimepack *tpp;
    struct dttimepack *tpp1;

    /*
     * Put the DAT drive in audio mode.
     */
    mtc.mt_op = MTAUD;
    mtc.mt_count = 1;
    if (ioctl(fd, MTIOCTOP, &mtc) < 0) {
        if (oserror() != EAGAIN) {
            perror("Couldn't put DAT into audio mode");
            return(0);
        }
    }

    /*
     * Now get some information about the DAT media
     * We'll use this momentarily.
     */
    if (ioctl(fd, MTIOCGET, &mtg) < 0) {
        perror("Couldn't issue MTIOCGET");
        return(0);
    }

    /*
     * rewind the tape.
     * Note that a rewind does not block. The rewind goes on while
     * the program continues. The first write will block until
     * the rewind completes.
     * We want the drive in audio mode here because BOT is different
     * between data and audio modes (data mode rewinds to logical
     * BOT, which is a little ways into the tape, and we want real
     * BOT).
     */
    if ((mtg.mt_erreg & (CT_AUD_MED >> 16)) == 0 &&
        (mtg.mt_dsreg & CT_BOT)) {
        /*
         * firmware bug workaround. Tapes which have previously
         * been written with non-audio data are considered "data
         * tapes." Their format differs from that of audio tapes.
         * If the tape in the drive  is considered a data tape, and
         * the drive thinks it's at BOT, we want to move it off
         * BOT so that the drive will really rewind it when we issue
         * MTREW. Can't do a read, since a read of a data tape in
         * audio mode will fail. The write will force the medium to
         * become an audio tape.
         */
    if (verbose)
            printf("fixing a data tape at BOT to be an audio tape at BOT\n");
        write(fd,&empty,sizeof(DTFRAME));
    }
    mtc.mt_op = MTREW;
    mtc.mt_count = 1;
    ioctl(fd, MTIOCTOP, &mtc);

    /*
     * Now we stick 100mm guard area on the start
     * of the tape.
     */
    bzero(&empty, sizeof(DTFRAME));
    empty.subcode.sid.pno1 = 0;
    empty.subcode.sid.pno2 = empty.subcode.sid.pno3 = 0xB;   /* BOT */
    empty.subcode.sid.ctrlid =  DTS_START;

    /*
     * Write all timecode packs, and indicate values in
     * all are invalid (but present!)
     */
    empty.subcode.mid.sampfreq = DT_FREQ44100;
    empty.subcode.packs[DTP_ATIME-1].id = DTP_ATIME;
    empty.subcode.packs[DTP_PTIME-1].id = DTP_PTIME;
    empty.subcode.packs[DTP_RTIME-1].id = DTP_RTIME;
    tpp = (struct dttimepack *)&empty.subcode.packs[DTP_ATIME-1];
    tcp = &tpp->tc;
    tcp->hhi = tcp->hlo = 0;
    tcp->mhi = tcp->mlo = tcp->shi = tcp->slo = tcp->fhi = tcp->flo = 0xA;
    tpp = (struct dttimepack *)&empty.subcode.packs[DTP_PTIME-1];
    tcp = &tpp->tc;
    tcp->hhi = tcp->hlo = 0;
    tcp->mhi = tcp->mlo = tcp->shi = tcp->slo = tcp->fhi = tcp->flo = 0xA;
    tpp = (struct dttimepack *)&empty.subcode.packs[DTP_RTIME-1];
    tcp = &tpp->tc;
    tcp->hhi = tcp->hlo = 0;
    tcp->mhi = tcp->mlo = tcp->shi = tcp->slo = tcp->fhi = tcp->flo = 0xA;
    empty.subcode.sid.numpacks = 7;

    /*
     * Write this dummy frame to the tape to make the leader
     * (i've always wondered how dummies became leaders...)
     */
    for (i = 0; i < 100; i++) {
        if ((write(fd, &empty, sizeof(DTFRAME))) < 0) {
            perror("tape write failed on leader");
            return 0;
        }
    }

    /*
     * initialize a dat frame
     */
    bzero(&dtf,sizeof(DTFRAME));

    /*
     * set up the rate bits
     */
    dtf.subcode.mid.sampfreq = DT_FREQ44100;

    /*
     * set up the subcode bits
     */
    tpp = (struct dttimepack *)&dtf.subcode.packs[DTP_ATIME-1];
    tpp->id = DTP_ATIME;

    tpp->index.dhi = 0;
    tpp->index.dlo = 1;
    tpp->pno1 = dtf.subcode.sid.pno1 = 0;
    tpp->pno2 = dtf.subcode.sid.pno2 = 0;
    tpp->pno3 = dtf.subcode.sid.pno3 = 1;
    tpp1 = (struct dttimepack *)&dtf.subcode.packs[DTP_PTIME-1];
    *tpp1 = *tpp;
    tpp1->id = DTP_PTIME;

    /* mark R-time as invalid, since we don't really want to write it */
    tpp = (struct dttimepack *)&empty.subcode.packs[DTP_RTIME-1];
    tpp->id = DTP_RTIME;
    tcp = &tpp->tc;
    tcp->hhi = tcp->hlo = 0;
    tcp->mhi = tcp->mlo = tcp->shi = tcp->slo = tcp->fhi = tcp->flo = 0xA;
    dtf.subcode.sid.ctrlid = 0;

    /* numpacks tells how many packs are written onto the DAT */
    dtf.subcode.sid.numpacks = 7;    /* ATIME is 2, PTIME is 1 */
}

int
prepare_dat_append(int fd)
{
    int i;
    struct mtop mtc;
    struct mtaudio mta;
    struct mtget mtg;
    struct dttimecode *tcp;
    struct dttimepack *tpp;
    struct dttimepack *tpp1;

    /*
     * Put the DAT drive in audio mode.
     */
    mtc.mt_op = MTAUD;
    mtc.mt_count = 1;
    if (ioctl(fd, MTIOCTOP, &mtc) < 0) {
        if (oserror() != EAGAIN) {
            perror("Couldn't put DAT into audio mode");
            return(0);
        }
    }

    /*
     * Now get some information about the DAT media
     * We'll use this momentarily.
     */
    if (ioctl(fd, MTIOCGET, &mtg) < 0) {
        perror("Couldn't issue MTIOCGET");
        return(0);
    }

    /*
     * rewind the tape.
     * Note that a rewind does not block. The rewind goes on while
     * the program continues. The first write will block until
     * the rewind completes.
     * We want the drive in audio mode here because BOT is different
     * between data and audio modes (data mode rewinds to logical
     * BOT, which is a little ways into the tape, and we want real
     * BOT).
     */
    if ((mtg.mt_erreg & (CT_AUD_MED >> 16)) == 0 &&
        (mtg.mt_dsreg & CT_BOT)) {
        /*
         * firmware bug workaround. Tapes which have previously
         * been written with non-audio data are considered "data
         * tapes." Their format differs from that of audio tapes.
         * If the tape in the drive  is considered a data tape, and
         * the drive thinks it's at BOT, we want to move it off
         * BOT so that the drive will really rewind it when we issue
         * MTREW. Can't do a read, since a read of a data tape in
         * audio mode will fail. The write will force the medium to
         * become an audio tape.
         */
    if (verbose)
            printf("fixing a data tape at BOT to be an audio tape at BOT\n");
        write(fd,&empty,sizeof(DTFRAME));
    }

    /* Setup leadout frame */

    bzero(&empty, sizeof(DTFRAME));
    empty.subcode.sid.pno1 = 0;
    empty.subcode.sid.pno2 = empty.subcode.sid.pno3 = 0xB;   /* BOT */
    empty.subcode.sid.ctrlid =  DTS_START;

    /*
     * Write all timecode packs, and indicate values in
     * all are invalid (but present!)
     */
    empty.subcode.mid.sampfreq = DT_FREQ44100;
    empty.subcode.packs[DTP_ATIME-1].id = DTP_ATIME;
    empty.subcode.packs[DTP_PTIME-1].id = DTP_PTIME;
    empty.subcode.packs[DTP_RTIME-1].id = DTP_RTIME;
    tpp = (struct dttimepack *)&empty.subcode.packs[DTP_ATIME-1];
    tcp = &tpp->tc;
    tcp->hhi = tcp->hlo = 0;
    tcp->mhi = tcp->mlo = tcp->shi = tcp->slo = tcp->fhi = tcp->flo = 0xA;
    tpp = (struct dttimepack *)&empty.subcode.packs[DTP_PTIME-1];
    tcp = &tpp->tc;
    tcp->hhi = tcp->hlo = 0;
    tcp->mhi = tcp->mlo = tcp->shi = tcp->slo = tcp->fhi = tcp->flo = 0xA;
    tpp = (struct dttimepack *)&empty.subcode.packs[DTP_RTIME-1];
    tcp = &tpp->tc;
    tcp->hhi = tcp->hlo = 0;
    tcp->mhi = tcp->mlo = tcp->shi = tcp->slo = tcp->fhi = tcp->flo = 0xA;
    empty.subcode.sid.numpacks = 7;

    /*
     * initialize a dat frame
     */
    bzero(&dtf,sizeof(DTFRAME));

    /*
     * set up the rate bits
     */
    dtf.subcode.mid.sampfreq = DT_FREQ44100;

    /*
     * set up the subcode bits
     */
    tpp = (struct dttimepack *)&dtf.subcode.packs[DTP_ATIME-1];
    tpp->id = DTP_ATIME;

    tpp->index.dhi = 0;
    tpp->index.dlo = 1;
    tpp->pno1 = dtf.subcode.sid.pno1 = 0;
    tpp->pno2 = dtf.subcode.sid.pno2 = 0;
    tpp->pno3 = dtf.subcode.sid.pno3 = 1;
    tpp1 = (struct dttimepack *)&dtf.subcode.packs[DTP_PTIME-1];
    *tpp1 = *tpp;
    tpp1->id = DTP_PTIME;

    /* mark R-time as invalid, since we don't really want to write it */
    tpp = (struct dttimepack *)&empty.subcode.packs[DTP_RTIME-1];
    tpp->id = DTP_RTIME;
    tcp = &tpp->tc;
    tcp->hhi = tcp->hlo = 0;
    tcp->mhi = tcp->mlo = tcp->shi = tcp->slo = tcp->fhi = tcp->flo = 0xA;
    dtf.subcode.sid.ctrlid = 0;

    /* numpacks tells how many packs are written onto the DAT */
    dtf.subcode.sid.numpacks = 7;    /* ATIME is 2, PTIME is 1 */

/* Find the end of the tape */

    seek_eot(fd);

    seek_to_last(fd);

    get_tape_pos(fd, 1);

    seek_to_last(fd);

/* Reset the index for a new track start */

    tpp->index.dhi = 0;
    tpp->index.dlo = 1;
    index = 0;
}

main(int argc, char **argv)
{
    CDSTATUS cdstatus;
    CDPARSER *cdparser;
    CDFRAME buf[NFRAMES];        /* CD input buffer */
    int status;
    int n = 1;
    int i;
    int excl_id;            /* id as returned by mediad */
    char uservfy[20];
    int maj, min;            /* dat firmware rel major&minor # */
    int nukeendmarker=0;
    int ret;
    int count;

    nukeendmarker = 0;
    verbose = 1;
    append=0;
    starttrack = -1;
    endtrack = -1;
    playaudio = 0;

    for(i=1;i<argc;i++)
        {
        if(strncmp(argv[i],"-h",2) == 0)
            {
		printf( "cdtodat - CDtoDAT v1.0 - Copy CD audio tracks to DAT via the SCSI bus\n");
		printf("\n");
		printf("SYNOPSIS\n");
		printf("\n");
		printf("    cdtodat -h\n");
		printf("    cdtodat [-f configfile] [-n] [-q] [-e track#]\n");
		printf("    cdtodat [-f configfile] [-a] [-q] [-p] [-e track#] ");
		printf("[starttrack] [endtrack]\n");
		printf("\n");
		printf("SWITCHES\n");
		printf("\n");
		printf("    -h  Show this help\n");
		printf("    -q  Be quiet during operations, normally we are VERY verbose\n");
		printf("    -n  Replace the special endmarker with tape leadout\n");
		printf("    -a  Append to tape AT the special end of recording track\n");
		printf("            number (299 by default)\n");
		printf("    -p  Play audio during copy\n");
		printf("    -e  Specify the special end of recording track number,\n");
		printf("            used for -n and -a\n");
		printf("    -f  Specify a configuration file, the configfile overrides\n");
		printf("            options that precede it on the command line\n");
            exit(0);
            }

        if(strcmp(argv[i],"-q") == 0) verbose = 0;
        if(strcmp(argv[i],"-n") == 0) nukeendmarker = 1;
        if(strcmp(argv[i],"-a") == 0) append=1;
        if(strcmp(argv[i],"-p") == 0) playaudio=1;

        if(strcmp(argv[i],"-e") == 0 && i+1 < argc)
            {
            aptrack = atoi(argv[i+1]);
            i++;
            continue;
            }

        if(strcmp(argv[i],"-f") == 0 && i+1 < argc)
            {
            configfile = argv[i+1];
            i++;
            continue;
            }

        if(tracklistlen == -1 && endtrack == -1 &&
           starttrack != -1 && isdigit(argv[i][0]))
            {
            endtrack = atoi(argv[i]);
            }

        if(tracklistlen == -1 && starttrack == -1 &&
           isdigit(argv[i][0]))
            {
            starttrack = atoi(argv[i]);
            }
        }

    if(configfile != NULL)
        {
        ret=parseconfigfile(configfile);
        if(ret == -1) fprintf(stderr,"Unable to process config file %s\n",
                              configfile);
        }

    fprintf(stderr,"Options in effect:\n");
    if(verbose) fprintf(stderr,"We are being Verbose\n");

    if(nukeendmarker)
        {
        fprintf(stderr,"We are removing the end marker (track 299)\n");
        }
    else
        {
        if(append)
            {
            fprintf(stderr,
                    "We are appending to tape ON TOP of track %d\n",
                    aptrack);
            }

        if(starttrack != -1)
            {
            fprintf(stderr,"We are starting with CD track %d\n", starttrack);
            }

        if(endtrack != -1)
            {
            fprintf(stderr,"We are ending with CD track %d\n", endtrack);
            }

        if(playaudio)
            {
            fprintf(stderr,"We are playing the audio during the copy\n");
            }
        }

    if(tracklistlen != -1)
        {
        fprintf(stderr,"Copying CD track(s): ");
        for(count=0;count<tracklistlen;count++)
            {
            fprintf(stderr,"%d ",tracklist[count]);
            }
        fprintf(stderr,"\n");
        }


    excl_id = mediad_get_exclusiveuse(datdev, "cdtodat");
    if(excl_id < 0) fprintf(stderr,"Unable to get exclusive access to DAT\n");

    if(nukeendmarker == 0)
        {
        /*
         * open the default CD drive. This will get exclusive access
         * from mediad.
         */
        cd = CDopen(0, "r");
        if (!cd) {
            fprintf(stderr,"Couldn't open CD-ROM drive\n");
            exit(-1);
        }

        cdparser = CDcreateparser();
        if (!cdparser) {
            fprintf(stderr,"Couldn't create CD parser\n");
            exit(-1);
        }

        /*
         * add callback functions to deal with events from the CD-ROM
         * here's why:
         *    a.   we do NOT read ptime from the CD. We
         *         calculate it ourselves, starting from 0 at the beginning
         *         of a program, and incrementing from the start of the first
         *         non-zero index. Thus we have no ptime callback.
         *      b.   we do NOT read atime from the CD. We calculate it
         *         ourselves. Thus we have no atime callback.
         *    c.   we read the control bits to pass through the copy
         *         prohibit bit. This is legally required. Don't remove it.
         *         (cd_control_func)
         *    d.   we transfer index and program number from CD to DAT.
         *         (cd_index_func, cd_pnum_func)
         *    e.   The cd_audio function transfers the audio data to a queue,
         *         accounting for the difference in frame size
         *           between the media.
         */

        CDaddcallback(cdparser,cd_control,(CDCALLBACKFUNC) cd_control_func, 0);
        CDaddcallback(cdparser, cd_catalog, (CDCALLBACKFUNC)  cd_cat_func, 0);
        CDaddcallback(cdparser, cd_index, (CDCALLBACKFUNC)  cd_index_func, 0);
        CDaddcallback(cdparser, cd_pnum, (CDCALLBACKFUNC) cd_pnum_func, 0);
        CDaddcallback(cdparser, cd_audio, (CDCALLBACKFUNC) cd_audio_func, 0);

        /*
         * initialize our idea of the state of the CD-ROM
         */
        CDgetstatus(cd, &cdstatus);

        /*
         * Make sure everything is kosher with the CD-ROM. Must
         * have an audio disc in before the program is run.
         */
        if (cdstatus.state == CD_NODISC) {
            fprintf(stderr,"No disc in CD-ROM drive\n");
            exit(-1);
        }
        else if (cdstatus.state == CD_CDROM) {
            fprintf(stderr,"Disc in CD-ROM drive is not an audio CD\n");
            exit(-1);
        }
        else if (cdstatus.state == CD_ERROR) {
            fprintf(stderr,"Error reading CD-ROM drive\n");
            exit(-1);
        }

        if(starttrack != -1)
            {
            if(CDseektrack(cd, starttrack) == -1)
                {
                fprintf(stderr,"Unable to start from track %d\n",starttrack);
                }
            else
                {
                starttrack--;
                }
            }
        } /* END Nukemarker == 0 */
    /*
     * Open the DAT drive
     */

    datfd = open(datdev, O_RDWR);

    if (datfd < 0)
        {
        perror("Could not open DAT drive:");
        exit(-1);
        }

    signal(SIGTERM, stopdat);
    signal(SIGINT, stopdat);
    signal(SIGKILL, stopdat);

    if (get_firmware_revision(datfd, &maj, &min) < 0) {
        fprintf(stderr,"Couldn't get DAT firmware revision\n");
        exit(-1);
    }
    /* pre-2.63 DAT drives don't work with audio */
    if (maj < 2 || (maj == 2 && min < 63)) {
        fprintf(stderr,"DAT firmware rev %d.%d is too old -- you must have 2.63 or greater\n",maj,min);
        exit(-1);
    }

    if(append == 1 || nukeendmarker == 1)
        {
        if (prepare_dat_append(datfd) < 0)
            {
        exit(-1);
            }
        }
    else
        {
        if (prepare_dat(datfd) < 0)
            {
        exit(-1);
            }
        }

    if(append || nukeendmarker)
        {
        printf("Is this frame the last valid frame (YES/NO):");
        gets(uservfy);
        if(strcasecmp(uservfy,"yes") != 0)
            {
            printf("Aborting execution\n");
            exit(-1);
            }
        }
    else
        {
        printf("COMPLETELY overwrite tape (YES/NO) ");
        gets(uservfy);
        if(strcasecmp(uservfy,"yes") != 0)
            {
            printf("Aborting execution\n");
            exit(-1);
            }
        }

    if(nukeendmarker) goto doleadout;

    sproc(write_DAT,PR_SALL,datfd);

    while(n)
        {
        if(abortrecording == 1) break;

        n = CDreadda(cd, buf, NFRAMES);

        if (n < 0)
            {
            fprintf(stderr,"Error reading CD-ROM drive\n");
            finish = 1;
            exit(-1);
            }

        for (i = 0; i < n; i++)
            {
            CDparseframe(cdparser, &buf[i]);
            }
        }

    fprintf(stderr,"Waiting for DAT to stop writing\n");

    finish = 1;
    wait(&status);            /* wait for child to complete */

    /*
     * now finish up the DAT tape with a lead-out section marking
     * EOT
     */
    if(!nukeendmarker) write_end_track(datfd);

doleadout:

    if (finish_dat(datfd) < 0) {
    exit(-1);
    }

    if(verbose) printf("DAT tape is complete\n");

}


void
myDTframetohmsf(unsigned long fr, int *h,int *m, int *s, int *f)
{
    unsigned long mrem,hrem;
    *h = fr / 120000;
    hrem = fr % 120000;
    *m = hrem / 2000;
    mrem = hrem % 2000;
    *s = (2+(mrem * 3)) / 100;
    *f = mrem - (100 * *s) / 3;
}

void
myDTframetotc(unsigned long fr, struct dttimecode* tc)
{
    int h,m,s,f;
    myDTframetohmsf(fr,&h,&m,&s,&f);
    /*
     * now set the msf fields in the timecode. These are in BCD;
     */
    tc->hhi = h / 10;
    tc->hlo = h % 10;
    tc->mhi = m / 10;
    tc->mlo = m % 10;
    tc->shi = s / 10;
    tc->slo = s % 10;
    tc->fhi = f / 10;
    tc->flo = f % 10;
}

int
get_tape_pos(int fd, int reseek)
{
    struct mtaudio mt_aud;
    struct mtget mt_get;
    struct dttimepack *tpp;
    struct dttimepack *tpp1;
    struct dttimecode *tcp;
    struct dttimecode *tcp1;
    int i;
    int need_writeback = 1;

    /*
     * if we just did a read, we don't need a new frame of data.
     * Otherwise, to figure out where we are, we must read a frame.
     * reseek set to 1 will force a read (and later re-seek to the
     * beginning of the frame it just read).
     */
    if (verbose) printf("Reading a frame...\n");
        if ((read(fd, &dtf, sizeof(DTFRAME))) < 0) {
            perror("tape read failed");
            return 0;
        }
    /*
     * get the program number from the frame
     */
    switch(dtf.subcode.sid.pno3) {
    case 0xa :
        /* unused program number -- ignore */
        break;
    case 0xb :
        /* bot */
        fprintf(stderr, "at BOT\n");
        return 0;
    case 0xe :
        /* eot */
        fprintf(stderr, "at EOT\n");
        return 0;
    default:
            program = DTpnotodec(dtf.subcode.sid.pno1,
            dtf.subcode.sid.pno2, dtf.subcode.sid.pno3);

        dtf.subcode.sid.pno1 = dtf.subcode.sid.pno1;
        dtf.subcode.sid.pno2 = dtf.subcode.sid.pno2;
        dtf.subcode.sid.pno3 = dtf.subcode.sid.pno3;
        break;
    }
    tpp1 = (struct dttimepack *)&dtf.subcode.packs[DTP_PTIME-1];
    if (DTtcvalid(&(tpp1->tc))) {
        tpp = (struct dttimepack *)&dtf.subcode.packs[DTP_PTIME-1];
    *tpp = *tpp1;
    }
    else {
    fprintf(stderr,"No valid ptime found at location -- setting ptime to 0\n");
        tpp = (struct dttimepack *)&dtf.subcode.packs[DTP_PTIME-1];
        tpp1 = (struct dttimepack *)&dtf.subcode.packs[DTP_PTIME-1];
    tpp->pno1 = tpp1->pno1;
    tpp->pno2 = tpp1->pno2;
    tpp->pno3 = tpp1->pno3;
    tpp->index = tpp1->index;
    bzero(&tpp->tc,sizeof(struct dttimecode));
    }
    if ((i = DTbcdtodec(tpp1->index.dhi,tpp1->index.dlo)) < 100) {
    index = i;
    }
    tpp1 = (struct dttimepack *)&dtf.subcode.packs[DTP_ATIME-1];
    if (DTtcvalid(&(tpp1->tc))) {
        tpp = (struct dttimepack *)&dtf.subcode.packs[DTP_ATIME-1];
    *tpp = *tpp1;
    }
    else {
    fprintf(stderr,"No valid atime found at location\n");
        tpp = (struct dttimepack *)&dtf.subcode.packs[DTP_PTIME-1];
        tpp1 = (struct dttimepack *)&dtf.subcode.packs[DTP_PTIME-1];
    tpp->pno1 = tpp1->pno1;
    tpp->pno2 = tpp1->pno2;
    tpp->pno3 = tpp1->pno3;
    tpp->index = tpp1->index;
    /*
     * If there's no valid A-time then we fail. We need valid
     * a-time, if not just for re-seeking to this frame.
     */
        return 0;
    }
    if ((i = DTbcdtodec(tpp1->index.dhi,tpp1->index.dlo)) < 100) {
    index = i;
    }
    if (verbose) {
        char *str="--:--:--:--";
        tpp = (struct dttimepack *)&dtf.subcode.packs[DTP_ATIME-1];
    DTtimetoa(str,&tpp->tc);
    printf("DAT pos (read frame): prog %d index %d atime %s",program,index,str);
        tpp = (struct dttimepack *)&dtf.subcode.packs[DTP_PTIME-1];
    DTtimetoa(str,&tpp->tc);
    printf(" ptime %s\n",str);
    }

    if (reseek == 0 || need_writeback) {
    /*
     * if we are not going to move the tape back to where it was,
     * at least adjust the time to reflect the correct timecode where
     * we are.
     */
        tpp = (struct dttimepack *)&dtf.subcode.packs[DTP_ATIME-1];
        tcp = &tpp->tc;
        DTinctime(tcp);
        /*
         * The DAT spec states that when index=0, ptime stays
         * at 0. We therefore only increment for a nonzero index.
         */
        if (index != 0) {
            tpp = (struct dttimepack *)&dtf.subcode.packs[DTP_PTIME-1];
            tcp = &tpp->tc;
            DTinctime(tcp);
        }
    }

    return 1;
}


int
seek_eot(int fd)
{
    struct mtget mt_get;
    int sawvalid = 0;
    struct mtaudio mt_aud;
    struct mtop mtc;
    int pt;
    int pn;
    char *s = "--:--:--:--";

    if (verbose) printf("Rewinding tape for safety\n");

    mtc.mt_op = MTREW;
    mtc.mt_count = 1;
    if(ioctl(fd, MTIOCTOP, &mtc) < 0){
        perror("DAT rewind failed");
        return 0;
    }

    if (verbose) {
    printf("Seeking to program %d\n\n",aptrack);
        fflush(stdout);
    }

    bzero(&mt_aud, sizeof(mt_aud));

    mt_aud.pno1 = aptrack / 100;
    pt = aptrack % 100;
    mt_aud.pno2 = pt / 10;
    mt_aud.pno3 = pt % 10;
    mt_aud.seektype = MTAUDPOSN_PROG;

    /*
     * start the seek.
     */
    if (ioctl(fd, MTSETAUDIO, &mt_aud) < 0) {
        perror("DAT seek to end failed");
        return 0;
    }

    /*
     * wait for the seek to complete.
     */

    do  {
        if(ioctl(fd, MTGETAUDIO, &mt_aud) < 0)
            {
            perror("DAT query failed");
            return 0;
            }

    pn = DTpnotodec(mt_aud.pno1, mt_aud.pno2, mt_aud.pno3);

        if(pn != aptrack) last_valid_atime = mt_aud.atime;

        if(verbose) {
        pn = DTpnotodec(mt_aud.pno1, mt_aud.pno2, mt_aud.pno3);

            if(pn != aptrack && pn > highprog) highprog = pn;

        DTtimetoa(s,&mt_aud.atime);
        fprintf(stderr,"DAT pos: prog %d atime %s",pn,s);
        DTtimetoa(s,&mt_aud.ptime);
        fprintf(stderr," ptime %s    \r",s);
            }

        if(ioctl(fd, MTIOCGET, &mt_get) < 0)
            {
            perror("DAT query failed");
            return 0;
            }

        } while(mt_get.mt_erreg & (CT_SEEKING >> 16));

    fprintf(stderr,"\nHigh Program %d\n",highprog);

    fprintf(stderr,"\n");
    return 1;
}

int
seek_to_last(int fd)
{
    struct mtget mt_get;
    int sawvalid = 0;
    struct mtaudio mt_aud;
    struct mtop mtc;
    int pt;
    int pn;
    char *s = "--:--:--:--";

    DTtimetoa(s,&last_valid_atime);
    fprintf(stderr,"\nSeeking to ATIME %s\n",s);

    bzero(&mt_aud, sizeof(mt_aud));
    mt_aud.atime = last_valid_atime;

    mt_aud.seektype = MTAUDPOSN_ABS;

    /*
     * start the seek.
     */
    if (ioctl(fd, MTSETAUDIO, &mt_aud) < 0) {
        perror("DAT seek to end failed");
        return 0;
    }

    /*
     * wait for the seek to complete.
     */

    do  {
        if(ioctl(fd, MTGETAUDIO, &mt_aud) < 0)
            {
            perror("DAT query failed");
            return 0;
            }

        if(verbose) {
        pn = DTpnotodec(mt_aud.pno1, mt_aud.pno2, mt_aud.pno3);

        DTtimetoa(s,&mt_aud.atime);
        fprintf(stderr,"DAT pos: prog %d atime %s",pn,s);
        DTtimetoa(s,&mt_aud.ptime);
        fprintf(stderr," ptime %s    \r",s);
            }

        if(ioctl(fd, MTIOCGET, &mt_get) < 0)
            {
            perror("DAT query failed");
            return 0;
            }

        } while(mt_get.mt_erreg & (CT_SEEKING >> 16));

fprintf(stderr,"\n");
return 1;
}

void
increment_time()
{
    struct dttimepack *tpp;
    struct dttimecode *tcp;

    tpp = (struct dttimepack *)&dtf.subcode.packs[DTP_ATIME-1];
    tcp = &tpp->tc;
    DTinctime(tcp);
    /*
     * The DAT spec states that when index=0, ptime stays
     * at 0. We therefore only increment for a nonzero index.
     */

    if (index != 0) {
        tpp = (struct dttimepack *)&dtf.subcode.packs[DTP_PTIME-1];
        tcp = &tpp->tc;
        DTinctime(tcp);
    }

    if (startidframes > 0) {
        startidframes--;
        if(startidframes==0) {
            /* shut off start ID bit after startidframes */
            dtf.subcode.sid.ctrlid &= ~DTS_START;
        }
    }
}


int
write_end_track(int fd)
{
    struct dttimecode *tcp;
    struct dttimepack *tpp;
    unsigned long i;

    if (verbose) {
        printf("Writing end marker (Program 299)\n");
    }

    /*
     * set up the subcode bits
     * first we do a-time
     */
    tpp = (struct dttimepack *)&dtf.subcode.packs[DTP_ATIME-1];

    tpp->pno1 = dtf.subcode.sid.pno1 = 2;
    tpp->pno2 = dtf.subcode.sid.pno2 = 9;
    tpp->pno3 = dtf.subcode.sid.pno3 = 9;

    /*
     * now we do p-time
     * we zero p-time when we change program numbers.
     * The DAT spec sez that program time is set to zero at start-of-
     * program, and increments throughout the program. The start-of-
     * program is where the first nonzero index occurs within the program.
     */
    tpp = (struct dttimepack *)&dtf.subcode.packs[DTP_PTIME-1];
    tcp = &tpp->tc;

    tcp->hhi = 0;
    tcp->hlo = 0;
    tcp->mhi = 0;
    tcp->mlo = 0;
    tcp->shi = 0;
    tcp->slo = 0;
    tcp->fhi = 0;
    tcp->flo = 0;
    tpp->pno1 = dtf.subcode.sid.pno1;
    tpp->pno2 = dtf.subcode.sid.pno2;
    tpp->pno3 = dtf.subcode.sid.pno3;

    index = 1;
    /* ...and we don't bother with r-time */

    /*
     * now we set it up so that the next 300 frames
     * have the start id bit set, indicating start-of-program.
     * DAT spec sez START must be set for 300 frames at start-of-program
     * and PRIORITYID must be set whenever program # is valid.
     */
    dtf.subcode.sid.ctrlid |= (DTS_PRIORITYID | DTS_START);
    startidframes = 300;

    /* set the audio in the template frame to zero (silence) */

    bzero(&dtf.audio,DTDA_DATASIZE);

    for (i = 0; i < 2000; i++) {
        fix_parity();
        if(verbose) fprintf(stderr,"Writing frame %d\r",i);
        if ((write(fd, &dtf, sizeof(DTFRAME))) < 0) {
            perror("tape write failed");
            return 0;
        }
        increment_time();
    }

  fprintf(stderr,"\n");

}

int parseconfigfile(char *filename)
{
char buffer[BUFSIZ];
char op[30];
int opval;
FILE *ifp;
int ret;
int count;

ifp=fopen(filename,"r");

if(ifp == NULL) return -1;

tracklistcount = 0;

while(fgets(buffer,BUFSIZ,ifp) != NULL)
    {
    if(buffer[0] == '#') continue;

    ret=sscanf(buffer,"%s %d",op, &opval);

    if(ret == 1)
        {
        if(strcasecmp(op,"quiet") == 0)        verbose = 0;
        if(strcasecmp(op,"playaudio") == 0)    playaudio = 1;
        if(strcasecmp(op,"append") == 0)       append = 1;
        }

    if(ret == 2)
        {
        if(strcasecmp(op,"eottrack") == 0)     aptrack = opval;
        if(strcasecmp(op,"startcdtrack") == 0) starttrack = opval;
        if(strcasecmp(op,"endcdtrack") == 0)   endtrack = opval;

        if(strcasecmp(op,"addcdtrack") == 0)
            {
            tracklist[tracklistcount] = opval;
            tracklistcount++;
            }
        }
    }

if(tracklistcount != 0)
    {
    tracklistlen = tracklistcount;
    tracklistcount = 0;
    starttrack = -1;
    endtrack = -1;
    }

fclose(ifp);
return 1;
}

