%token LEADIN LEADOUT BOT EOT RATE COPY INTERVAL PROGRAM INDEX PAUSE SILENCE FRAMES FREAD FWRITE SEEK INTEGER BOOLEAN PROGNUM ATIME PTIME COLON WHERE WHERENOT READFRAMES READFRAME QUIT STRING TESTPAT COMMENT ABORT WATCHSEEK SHADOW CHECK PRINT
%{
extern int dat;
static int h,m,s,f;
static int p;
static int i;
extern char last_string[];
%}
%%
script:	command |
	script command 
	;
command:	leadin | leadout | rate | copy | program | index | pause | interval | silence | seek | where | wherenot | atime | ptime | readframes | readframe | quit | write | read | testpat | comment | abort | watchseek | shadow | check | print;

quit: QUIT
	{
		quit();
	}
	;
abort: ABORT
	{
		mt_abort();
	}
	;
where:	WHERE
	{
		get_tape_pos(dat,0);
	}
	;
wherenot:	WHERENOT
	{
		read_position(dat);
	}
	;
readframe:	READFRAME
	{
		get_tape_pos(dat,1);
	}
	;
readframes:	READFRAMES INTEGER
	{
		for (i = 0 ; i < yylval; i++) {
			get_tape_pos(dat,1);
		}
	}
	;
leadin:	LEADIN
	{
		write_bot(dat);
	}
	;
leadout: LEADOUT
	{
		write_eot(dat);
	}
	;
rate:	RATE INTEGER
	{
		set_rate(yylval);
	}
	;
interval:	INTERVAL INTEGER
	{
		set_interval(yylval);
	}
	;
copy:	COPY BOOLEAN
	{
		copy_prohibit(yylval);
	}
	;
watchseek:	WATCHSEEK BOOLEAN
	{
		watch_seek(yylval);
	}
	;

print: PRINT
	{
		print("");
	}
	| PRINT STRING 
	{
		print(last_string);
	}
	;

shadow: SHADOW STRING
	{
		shadow(last_string);
	}
	;

read:	FREAD STRING prog
	{
		read_file(dat,last_string,p,0,0,0,0);
	}
	| FREAD STRING FRAMES INTEGER
	{
		read_frames(dat,last_string,yylval);
	}
	| FREAD STRING timecode
	{
		read_file(dat,last_string,0,h,m,s,f);
	}
	;

write:	FWRITE STRING
	{
		write_file(dat,last_string);
	}
	;

program: PROGRAM
	{
		inc_program();
	}
	| PROGRAM INTEGER 
	{
		set_program(yylval);
	}
	;

index: INDEX
	{
		inc_index();
	}
	| INDEX INTEGER
	{
		set_index(yylval);
	}
	;

pause: PAUSE
	{
		pause();
	}
	|
	PAUSE INTEGER
	{
		pause_time(yylval);
	}
	;

testpat: TESTPAT INTEGER
	{
		write_testpat(dat,yylval);
	}
	;

silence: SILENCE INTEGER
	{
		write_silence(dat,yylval);
	}
	;

atime: ATIME timecode 
	{
		set_atime(h,m,s,f);
	}
	;
ptime: PTIME timecode
	{
		set_ptime(h,m,s,f);
	}
	;
prog: PROGNUM INTEGER { p = yylval; };

seek: SEEK ATIME timecode 
	{ 
		seek_time(dat,1,h,m,s,f);
	}
      | SEEK PTIME timecode
	{
		seek_time(dat,0,h,m,s,f);
	}
      | SEEK prog 
	{
		seek_prog(dat,p);
	}
      | SEEK BOT
	{
		seek_bot(dat);
	}
      | SEEK EOT
	{
		seek_eot(dat);
	}
      ;

check: CHECK ATIME timecode 
	{ 
		check_time(dat,1,h,m,s,f);
	}
      | CHECK PTIME timecode
	{
		check_time(dat,0,h,m,s,f);
	}
      | CHECK prog 
	{
		check_prog(dat,p);
	}
      | CHECK BOT
	{
		check_bot(dat);
	}
      | CHECK EOT
	{
		check_eot(dat);
	}
      ;

hours: INTEGER { h = yylval; };
minutes: INTEGER { m = yylval; };
seconds: INTEGER { s = yylval; };
frames: INTEGER { f = yylval; };

timecode: hours COLON minutes COLON seconds COLON frames;

comment: COMMENT ;
